# -*- coding: utf-8 -*-
"""Simple command line based habits tracker."""

import click
import logging

from datetime import datetime, timedelta
from os import path, mkdir
from sys import float_info

import habito.models as models

logger = logging.getLogger("habito")
database_name = path.join(click.get_app_dir("habito"), "habito.db")
TERMINAL_WIDTH, TERMINAL_HEIGHT = click.get_terminal_size()

TICK = u"\u25A0"    # tick - 2713, black square - 25A0, 25AA, 25AF
CROSS = u"\u25A1"   # cross - 2717, white square - 25A1, 25AB, 25AE
PARTIAL = u"\u25A0"    # tick - 2713, black square - 25A0, 25AA, 25AF


@click.group()
def cli():
    """Habito - a simple command line habit tracker."""
    if not path.exists(click.get_app_dir("habito")):
        mkdir(click.get_app_dir("habito"))
    models.setup(database_name)


@cli.command()
@click.option("-l", is_flag=True, help="Long listing with date and quantum.")
def list(l):
    """List all tracked habits."""
    from terminaltables import SingleTable
    from textwrap import wrap

    nr_of_dates = TERMINAL_WIDTH // 10 - 4
    if nr_of_dates < 1:
        logger.debug("list: Actual terminal width = {0}.".format(click.get_terminal_size()[0]))
        logger.debug("list: Observed terminal width = {0}.".format(TERMINAL_WIDTH))
        click.echo("Your terminal window is too small. Please make it wider and try again")
        raise SystemExit(1)

    table_title = ["Habit", "Goal", "Streak"]
    minimal = not l
    if minimal:
        table_title.append("Activities")
    else:
        for d in range(0, nr_of_dates):
            date_mod = datetime.today() - timedelta(days=d)
            table_title.append("{0}/{1}".format(date_mod.month, date_mod.day))

    table_rows = [table_title]
    for habit_data in models.get_daily_activities(nr_of_dates):
        habit = habit_data[0]
        habit_row = [str(habit.id) + ": " + habit.name, str(habit.quantum)]
        progress = ""
        for daily_data in habit_data[1]:
            column_text = CROSS
            quanta = daily_data[1]

            if quanta is not None:
                column_text = click.style(PARTIAL)
                if quanta >= habit.quantum:
                    column_text = click.style(TICK, fg="green")
            if minimal:
                progress += column_text + " "
            else:
                habit_row.append(quanta)
        if minimal:
            habit_row.append(progress)

        current_streak = habit.summary.get().get_streak()
        habit_row.insert(2, current_streak)
        table_rows.append(habit_row)

    table = SingleTable(table_rows)

    max_col_width = table.column_max_width(0)
    max_col_width = max_col_width if max_col_width > 0 else 20

    for r in table_rows:
        r[0] = '\n'.join(wrap(r[0], max_col_width))

    click.echo(table.table)


@cli.command()
@click.argument("name", nargs=-1)
@click.argument("quantum", type=click.FLOAT)
@click.option("--units", "-u", default="units", help="Units of data.")
def add(name, quantum, units):
    """Add a habit."""
    habit_name = ' '.join(name)
    models.Habit.add(name=habit_name,
                     created_date=datetime.now(),
                     quantum=quantum,
                     units=units,
                     magica="")

    msg_unit = click.style("{0} {1}".format(quantum, units), fg='green')
    msg_name = click.style("{0}".format(habit_name), fg='green')
    click.echo("You have commited to {0} of {1} every day!"
               .format(msg_unit, msg_name))


@cli.command()
@click.argument("id", type=click.INT)
@click.option('--name', '-n', help="The new name (leave empty to leave unchanged)")
@click.option('--quantum', '-q', help="The new quantum (leave empty to leave unchanged)", type=click.FLOAT)
def edit(id, name, quantum):
    """Edit a habit."""
    try:
        habit = models.Habit.get(models.Habit.id == id)
    except models.Habit.DoesNotExist:
        click.echo("The habit you're trying to edit does not exist!")
        raise SystemExit(1)
    habit.name = name.strip() or habit.name
    habit.quantum = quantum or habit.quantum
    habit.save()
    click.echo("Habit with id {} has been saved with name: {} and quantum: {}".format(id, habit.name, habit.quantum))


@cli.command()
@click.argument("id", type=click.INT)
@click.option("--keeplogs", is_flag=True, default=False,
              help="Preserve activity logs for the habit.")
def delete(id, keeplogs):
    """Delete a habit."""
    try:
        habit = models.Habit.get(models.Habit.id == id)
    except models.Habit.DoesNotExist:
        click.echo("The habit you want to remove does not seem to exist!")
        raise SystemExit(1)
    confirm = click.confirm("Are you sure you want to delete habit"
                            " {}: {} (this cannot be undone!)"
                            .format(habit.id, habit.name))
    if confirm:
        click.echo("Habit {}: {} has been deleted!".format(habit.id, habit.name))
        if not keeplogs:
            models.Activity.delete().where(models.Activity.for_habit ==
                                           habit.id).execute()
            models.Summary.delete().where(models.Summary.for_habit ==
                                          habit.id).execute()
            habit.delete_instance()
        else:
            habit.active = False
            habit.save()
    else:
        click.echo("Habit {}: {} has not been deleted!".format(habit.id, habit.name))


@cli.command()
@click.argument("name", nargs=-1)
@click.option("--review", "-r", is_flag=True,
              help="Update activity for all tracked habits.")
@click.option("--date", "-d",
              help="Date of activity in mm/dd format (defaults to today).",
              default=datetime.now().strftime("%m/%d"))
@click.option("--quantum", "-q", type=float, help="Progress for the day.")
def checkin(name, review, date, quantum):
    """Commit progress for a habit."""
    # Set a date for this checkin. Use past year if month/day is in future
    query = ' '.join(name)
    date = date.strip()
    d = datetime.strptime(date, "%m/%d").replace(year=datetime.now().year)
    update_date = d if d < datetime.now() else d.replace(year=d.year-1)
    update_date_str = update_date.strftime("%a %b %d %Y")

    def print_header(date_str):
        header = "Please update progress of habits for {0}:"
        click.echo(header.format(click.style(date_str, fg='green')))

    def get_quantum(habit, required=True):
        msg = "  - {0} (Goal: {1} {2})"

        # Keep prompting until we have a value if required is True
        value = None if required else float_info.max
        q = click.prompt(msg.format(habit.name, habit.quantum, habit.units),
                         type=float,
                         show_default=False,
                         default=value)
        if not required and q == value:
            return None
        return q

    def update_activity(habit, quantum, date):
        # Create an activity for this checkin
        activity = models.Activity.create(for_habit=habit,
                                          quantum=quantum,
                                          update_date=update_date)

        # Update streak for the habit
        models.Summary.update_streak(habit)
        return activity

    # Review mode: iterate through all habits
    if review:
        print_header(update_date_str)
        click.echo("(Press `enter` if you'd like to skip update for a habit.)")
        for h in models.Habit.all_active():
            q = get_quantum(h, required=False)
            if q is not None:
                update_activity(h, q, update_date)
        return

    # Non review mode: checkin a single habit
    if query == "":
        click.echo("No habit specified, no progress updated.")
        click.echo("Try 'habito checkin <habit_name>'?")
        return
    habits = models.Habit.all_active().where(models.Habit.name.regexp(query))
    if habits.count() == 0:
        error = "No habit matched the name '{0}'.".format(query)
        click.secho(error, fg='red')
        return
    elif habits.count() > 1:
        error = ("More than one habits matched the name '{0}'. "
                 "Don't know which to update.").format(query)
        click.secho(error, fg='red')
        for h in habits:
            click.echo("- {0}".format(h.name))
        click.echo("Try a different filter, or `habito checkin --review`.")
        return

    # Now add the activity for the chosen habit
    habit = habits[0]
    if quantum is None:
        print_header(update_date_str)
        quantum = get_quantum(habit, required=True)
    activity = update_activity(habit, quantum, update_date)

    act_msg = click.style("{0} {1}".format(activity.quantum, habit.units),
                          fg='green')
    act_date = click.style(update_date_str, fg='green')
    habit_msg = click.style("{0}".format(habit.name), fg='green')
    click.echo("Added {0} to habit {1} for {2}.".format(act_msg, habit_msg, act_date))


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
"""Models for habito."""

import logging
from datetime import datetime, timedelta
from peewee import *    # noqa
from playhouse import reflection
from playhouse.sqlite_ext import SqliteExtDatabase
from playhouse.migrate import SqliteMigrator, migrate

DB_VERSION = 2
db = SqliteExtDatabase(None, pragmas=(('foreign_keys', 'on'),),
                       regexp_function=True)
logger = logging.getLogger("habito.models")


def setup(name):
    """Set up the database."""
    db.init(name)
    db.connect()
    Migration(db).execute()


def get_activities(days):
    """Get activities of habits for specified days.

    Args:
        days (int): Number of days of activities to fetch.

    Returns:
        List of habits with the activities.

    """
    if days < 0:
        raise ValueError("Days should be a positive integer.")

    from_date = datetime.now() - timedelta(days=days)
    habits = Habit.all_active()
    activities = Activity.select()\
        .where(Activity.update_date > from_date)\
        .order_by(Activity.update_date.desc())

    habits_with_activities = prefetch(habits, activities)
    return habits_with_activities


def get_daily_activities(days):
    """Get activities for all habits grouped by day for specified days.

    Args:
        days (int): Number of days of activities to fetch.

    Returns:
        Tuple of habit and list of daily activities. Daily activities are the
        sum of all activities for the habit for the day.

        E.g. [(habit, [(day1, activity), (day2, activity)..]), ..]

    """
    habits_with_activities = get_activities(days)
    daily_habits = []
    for habit in habits_with_activities:
        habit_data = []

        activity_index = 0
        activities = habit.activities
        for day in range(0, days):
            quanta = 0.0
            if activity_index < len(activities):
                a = activities[activity_index]
            else:
                a = None

            # Assumes the data is sorted by date
            for_date = datetime.today() - timedelta(days=day)
            if a is None or a.update_date.date() != for_date.date():
                # We didn't find any data for the day
                quanta = None
            else:
                # Sum all activities for a day
                while a.update_date.date() == for_date.date():
                    quanta += a.quantum
                    activity_index += 1
                    if activity_index >= len(activities):
                        break
                    a = activities[activity_index]

            # Save the data for day
            habit_data.append((day, quanta))

        daily_habits.append((habit, habit_data))

    return daily_habits


class BaseModel(Model):
    """Base model class for Habito."""

    class Meta:
        """Meta class for the model."""

        database = db


class Config(BaseModel):
    """Database configuration.

    Attributes:
        name (str): Name of the setting
        value (str): Value

    """

    name = CharField(unique=True)
    value = CharField()


class Habit(BaseModel):
    """Represents a single habit.

    Attributes:
        name (str): Description of the habit.
        created_date (datetime): Date on which the habit was added.
        quantum (float): Amount for the habit.
        frequency (int): Data input frequency in numbers of days. (Default: 1)
        units (str): Units of the quantum.
        magica (str): Why is this habit interesting?
        active (bool): True if the habit is active

    """

    name = CharField()
    created_date = DateField(default=datetime.now())
    frequency = IntegerField(default=1)
    quantum = DoubleField()
    units = CharField()
    magica = TextField()
    active = BooleanField(default=True)

    @classmethod
    def add(cls, **query):
        """Add a habit.

        Args:
            query (kwargs): List of fields and values.

        Returns:
            A Habit.

        """
        habit = cls.create(**query)
        Summary.create(for_habit=habit,
                       target=0,
                       target_date=datetime.now(),
                       streak=0)
        return habit

    @classmethod
    def all_active(cls):
        """Add a habit.

        Returns:
            All active habits.

        """
        return cls.select().where(Habit.active)


class Activity(BaseModel):
    """Updates for a Habit.

    Attributes:
        for_habit (int): Id of the Habit. Foreign key.
        update_date (date): Date time of the update.
        quantum (float): Amount for the habit.

    """

    for_habit = ForeignKeyField(Habit, related_name="activities",
                                index=True)
    quantum = DoubleField()
    update_date = DateTimeField(default=datetime.now())


class Summary(BaseModel):
    """Continuous metrics for a Habit.

    Attributes:
        for_habit (int): Id of the Habit. Foreign key.
        target (float): A target for the Habit. Computed from the quantum.
        target_date (date): Date for the target.
        streak (int): Current streak for a habit. Continuous activity
        constitutes a streak.

    """

    for_habit = ForeignKeyField(Habit, related_name="summary",
                                index=True)
    target = DoubleField()
    target_date = DateField()
    streak = IntegerField(default=0)

    @classmethod
    def update_streak(cls, habit):
        """Update streak for a habit.

        Args:
            habit (Habit): Habit to update.
        """
        # Check-in now supports past updates for upto 365 days
        # TODO need an index on date, this query will scan entire table
        summary = cls.get(for_habit=habit)
        activities = list((Activity
                           .select(Activity.update_date,
                                   fn.SUM(Activity.quantum).alias('total_quantum'))
                           .where(Activity.for_habit == habit)
                           .group_by(fn.date(Activity.update_date))
                           .order_by(Activity.update_date.desc())))

        if len(activities) == 0:
            return summary

        last_update = activities[0].update_date
        streak = 1
        for activity in activities[1:]:
            diff = last_update - activity.update_date
            if diff.days > 1 or activity.total_quantum < habit.quantum:
                break
            last_update = activity.update_date
            streak += 1

        summary.streak = streak
        summary.save()
        return summary

    def get_streak(self):
        """Humanize a streak to include days."""
        streak = str(self.streak) + " day"
        if self.streak != 1:
            streak += "s"
        return streak


class Migration:
    """Migrations for habito database."""

    # Error codes for the migrations
    error_codes = {-1: "not run",
                   0: "success",
                   1: "generic failure"}

    def __init__(self, database):
        """Create an instance of the migration.

        Args:
            database: database instance
        """
        self._db = database

    def get_version(self):
        """Get the database version.

        DB version is 0 if database is new.
        DB version is 1 if tables exist, but Config table is not available
        DB version is the actual version otherwise.

        Args:
            database (Database): peewee database instance

        Returns:
            Database version (int).

        """
        # Return version 1 if other tables (excluding Config) exist
        try:
            tables = reflection.introspect(self._db).model_names
            if len(tables) == 0:
                return 0
            if "config" not in tables:
                return 1
            version = int(Config.get(Config.name == "version").value)
        except Config.DoesNotExist:
            # Config table exists but version is not present
            version = 1
        return version

    def execute(self, list_only=False):
        """Migrate the database schema to latest version."""
        # It's a new database if `act_ver=0`
        # Set current version to 0, we will only run migration_0
        act_ver = self.get_version()
        cur_ver = 0 if act_ver == 0 else DB_VERSION
        logger.debug("DB version: actual = {0}, current = {1}"
                     .format(act_ver, cur_ver))

        if cur_ver != 0 and act_ver == cur_ver:
            logger.debug("DB versions are same. Skip migration.")
            return {}

        def get_migration(version):
            return self.__getattribute__("_migration_{}".format(version))

        m = {x: (get_migration(x), -1) for x in range(act_ver, cur_ver+1)}
        if list_only:
            # List the migration and status without running
            return {f[0]: f[1][1] for f in m.items()}

        # Run the migrations and report their status
        return {f[0]: f[1][0]() for f in m.items()}

    def _migration_0(self):
        """Set latest state of the database schema."""
        self._db.create_tables([Config, Habit, Activity, Summary], safe=True)
        Config.create(name="version", value=str(DB_VERSION))
        return 0

    def _migration_1(self):
        """Apply migration #1."""
        tables = reflection.introspect(self._db).model_names
        migrator = SqliteMigrator(self._db)
        if "habitmodel" in tables and "activitymodel" in tables:
            with self._db.transaction():
                migrate(
                    migrator.rename_table("habitmodel", "habit"),
                    migrator.rename_table("activitymodel", "activity"))
            logger.debug("Migration #1: Renamed habit, activity tables.")

        # Create new tables
        self._db.create_tables([Config, Summary], safe=True)
        logger.debug("Migration #1: Created tables.")

        # Set DB version
        Config.insert(name="version", value="1").on_conflict("replace").execute()
        logger.debug("Migration #1: DB version updated to 1.")

        # Update summaries
        for h in Habit.select():
            activities = Activity.select()\
                                 .where(Activity.for_habit == h)\
                                 .order_by(Activity.update_date.asc())
            streak = 0
            if len(activities) != 0:
                last_date = activities[0].update_date
                for a in activities:
                    delta = last_date - a.update_date
                    if abs(delta.days) > 1:
                        break
                    streak += 1
                    last_date = a.update_date

            # Update summary for the habit
            s = Summary.get_or_create(for_habit=h, target=0,
                                      target_date=h.created_date)
            s[0].streak = streak
            s[0].save()
        logger.debug("Migration #1: Summary updated for habits.")
        return 0

    def _migration_2(self):
        """Apply migration #2.

        This is a dummy migration step.
        """
        # Set DB version
        Config.insert(name="version", value="2").on_conflict("replace").execute()
        logger.debug("Migration #2: DB version updated to 2.")
        return 0



"""Setup for Habito app."""

import os
import sys
from setuptools import setup, find_packages, Command
from shutil import rmtree
from codecs import open
from os import path

# Package meta-data.
NAME = "habito"
DESCRIPTION = "Simple command line habits tracker"
URL = "https://github.com/codito/habito"
EMAIL = "arun@codito.in"
AUTHOR = "Arun Mahapatra"
# VERSION = (1, 0, a2)

# Dependencies required for execution
REQUIRED = ["click", "peewee>=3.0.15", "terminaltables"]

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Print things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        """Initialize command options."""
        pass

    def finalize_options(self):
        """Finalize command options."""
        pass

    def run(self):
        """Run the upload command."""
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal"
                  .format(sys.executable))

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        sys.exit()


setup(
    name=NAME,
    version="1.0a5",    # FIXME semantic versioning
    description=DESCRIPTION,
    long_description=long_description,
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,

    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    entry_points={
        "console_scripts": ["habito=habito.habito:cli"],
    },
    install_requires=REQUIRED,

    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",

        "Environment :: Console",

        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",

        "License :: OSI Approved :: MIT License",

        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: POSIX :: Linux",

        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",

        "Topic :: Utilities",
    ],
    keywords="habits goals track tracking quantified self",
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
)
